import os
import traceback

from libs.message_handler import BaseMessageHandler
from libs.message import ContentType

from libs import logs
from libs.message import ProjectCommand
from project_manager import files, projects
from project_manager.interfaces import FileManagerMessageParams

log = logs.get_logger(__name__)


class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue, user_space=None):
        super(MessageHandler, self).__init__(p2n_queue, user_space)

    def handle_message(self, message):
        log.info('Handle FileExplorer message: %s' % message)
        try:
            # metadata = message.metadata
            # if 'path' in metadata.keys():
            #     # avoid creating `./` when the path is empty
            #     if metadata['path'] == "":
            #         norm_path = metadata['path']
            #     else:
            #         norm_path = os.path.normpath(metadata['path'])
            # if 'project_path' in metadata.keys():
            #     norm_project_path = os.path.normpath(metadata['project_path'])

            messageParams = FileManagerMessageParams(message.metadata)
            output = None
            type = None
            if message.command_name == ProjectCommand.list_dir:
                output = files.list_dir(
                    messageParams.norm_project_path, messageParams.norm_path)
                type = ContentType.DIR_LIST
            elif message.command_name == ProjectCommand.create_file:
                files.create_file(
                    messageParams.norm_project_path, messageParams.norm_path)
                output = projects.open_file(
                    messageParams.path, messageParams.open_order)
                type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.create_folder:
                file_metadata = files.create_folder(
                    messageParams.norm_project_path, messageParams.norm_path)
                if file_metadata != None:
                    output = file_metadata.path
                    type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.delete:
                file_metadata = files.delete(messageParams.norm_project_path,
                                messageParams.norm_path)

                # Delete state file
                state_file_path = files.get_state_path(
                    messageParams.norm_project_path, messageParams.norm_path)
                if os.path.exists(state_file_path):
                    files.delete(messageParams.norm_project_path,
                                 state_file_path)

                if messageParams.is_file:
                    output = projects.close_file(
                        messageParams.path, messageParams.open_order)
                else:  # TODO: handle the case where a dir is deleted and deleted files were opened
                    output = projects.get_open_files()
                type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.list_projects:
                output = projects.list_projects()
                type = ContentType.PROJECT_LIST

            # create reply message
            message.type = type
            message.content = output
            message.error = False
            self._send_to_node(message)

        except:
            trace = traceback.format_exc()
            log.error("%s" % (trace))
            error_message = MessageHandler._create_error_message(
                message.webapp_endpoint, trace, message.command_name)
            self._send_to_node(error_message)



import traceback
import simplejson as json
from libs.message import ContentType, Message, SubContentType
from libs import logs
# from server.python.libs.message import WebappEndpoint
from user_space.user_space import BaseKernel, IPythonUserSpace
from user_space.ipython.constants import IPythonConstants as IPythonConstants

log = logs.get_logger(__name__)


class BaseMessageHandler:
    def __init__(self, p2n_queue, user_space=None):
        self.p2n_queue = p2n_queue
        if user_space == None:
            self.user_space = IPythonUserSpace(BaseKernel())
        else:
            self.user_space = user_space

    @staticmethod
    def _is_execute_result(header) -> bool:
        return header['msg_type'] == IPythonConstants.MessageType.EXECUTE_RESULT

    @staticmethod
    def _is_execute_reply(header) -> bool:
        return header['msg_type'] == IPythonConstants.MessageType.EXECUTE_REPLY

    @staticmethod
    def _is_stream_result(header) -> bool:
        return header['msg_type'] == IPythonConstants.MessageType.STREAM

    @staticmethod
    def _is_display_data_result(header) -> bool:
        return header['msg_type'] == IPythonConstants.MessageType.DISPLAY_DATA

    @staticmethod
    def _is_error_message(header) -> bool:
        return header['msg_type'] == IPythonConstants.MessageType.ERROR
        
    ## TODO: this needs to be designed #
    # @staticmethod
    # def get_execute_result(messages):
    #     """
    #         Get result from list of messages are responsed by IPython kernel
    #         For result type rather than 'application/json' we have to convert the output
    #         to the original object before sending to client because we already json.dumps
    #         them inside ipython. A better way to handle this is to output 'application/json' 
    #         instead. That will be done later.
    #     """
    #     result = None
    #     for message in messages:
    #         log.info('Message type %s' % message['header']['msg_type'])
    #         if message['header']['msg_type'] == IPythonConstants.MessageType.EXECUTE_RESULT:
    #             if message['content']['data']['text/plain'] is not None:
    #                 result = message['content']['data']['text/plain']
    #                 result = json.loads(result)
    #         elif message['header']['msg_type'] == IPythonConstants.MessageType.STREAM:
    #             # log.info('Stream result: %s' % result)
    #             if 'text' in message['content']:
    #                 result = message['content']['text']
    #                 result = json.loads(result)
    #         elif message['header']['msg_type'] == IPythonConstants.MessageType.DISPLAY_DATA:
    #             if SubContentType.APPLICATION_PLOTLY in message['content']['data']:
    #                 result = message['content']['data'][SubContentType.APPLICATION_PLOTLY]
    #     return result

    @staticmethod
    def get_execute_result(message):
        """
            Get result from list of messages are responsed by IPython kernel
            For result type rather than 'application/json' we have to convert the output
            to the original object before sending to client because we already json.dumps
            them inside ipython. A better way to handle this is to output 'application/json' 
            instead. That will be done later.
        """
        result = None
        log.info('Message type %s' % message.header['msg_type'])
        if message.header['msg_type'] == IPythonConstants.MessageType.EXECUTE_RESULT:
            if message.content['data']['text/plain'] is not None:
                result = message.content['data']['text/plain']
                result = json.loads(result)
        # elif message['header']['msg_type'] == IPythonConstants.MessageType.STREAM:
        #     # log.info('Stream result: %s' % result)
        #     if 'text' in message['content']:
        #         result = message['content']['text']
        #         result = json.loads(result)
        elif message.header['msg_type'] == IPythonConstants.MessageType.DISPLAY_DATA:
            if SubContentType.APPLICATION_PLOTLY in message.content['data']:
                result = message.content['data'][SubContentType.APPLICATION_PLOTLY]
        return result

    @staticmethod
    def _create_error_message(webapp_endpoint, trace, command_name=None, metadata=None):
        return Message(**{
            "webapp_endpoint": webapp_endpoint,
            "command_name": command_name,
            "type": ContentType.STRING,
            "content": trace,
            "error": True,
            "metadata": metadata,
        })

    def _send_to_node(self, message: Message):
        # the current way of communicate with node server is through stdout with a json string
        log.info("Send to node server: %s %s %s %s %s" %
                 (message.webapp_endpoint, message.command_name, message.type, message.sub_type, message.metadata))
        BaseMessageHandler.send_message(self.p2n_queue, message)

    @staticmethod
    def send_message(channel, message: Message):
        channel.send(message.toJSON())

    def handle_message(self, message):
        ''' `ext_globals` is the user namespace where the user executes their command'''
        raise "Abstract function must be implemented by subclass"

    def shutdown(self):
        self.user_space.shutdown_executor()
        pass



import re
import requests
import base64
import collections
import itertools
import sys
import traceback
import simplejson as json

from libs.message_handler import BaseMessageHandler
from libs.message import ContentType, DFManagerCommand, SubContentType
from libs.json_serializable import ipython_internal_output, JsonSerializable
from cnextlib.mime_types import CnextMimeType

from libs import logs
from libs.message import Message, WebappEndpoint
from user_space.ipython.constants import IPythonInteral, IPythonConstants, IpythonResultMessage
from user_space.ipython.kernel import IPythonKernel
from user_space.user_space import ExecutionMode
log = logs.get_logger(__name__)


def total_size(o, handlers={}, verbose=False):
    """ Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}

    """
    def dict_handler(d): return itertools.chain.from_iterable(d.items())
    all_handlers = {tuple: iter,
                    list: iter,
                    collections.deque: iter,
                    dict: dict_handler,
                    set: iter,
                    frozenset: iter,
                    }
    all_handlers.update(handlers)     # user handlers take precedence
    seen = set()                      # track which object id's have already been seen
    # estimate sizeof object without __sizeof__
    default_size = sys.getsizeof(0)

    def sizeof(o):
        if id(o) in seen:       # do not double count the same object
            return 0
        seen.add(id(o))
        s = sys.getsizeof(o, default_size)

        if verbose:
            print(s, type(o), repr(o), file=sys.stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)


class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue,  user_space=None):
        super(MessageHandler, self).__init__(p2n_queue, user_space)

    def _get_file_content(self, file_path, mime_type):
        with open(file_path, 'rb') as file:
            return file.read()

    def _is_jsonable(self, obj):
        try:
            json.dumps(obj)
            return True
        except (TypeError, OverflowError):
            return False

    def _convert_to_str_if_not_jsonable(self, df):
        for c in df.columns:
            for r in df.index:
                if not self._is_jsonable(df.at[r, c]):
                    df.at[r, c] = str(df.at[r, c])
        return df

    def _create_table_data(self, df_id, df):
        ''' The tableData will be the same as the original `df` except for two exceptions
            1. if column type is datatime64[ns], the data will be reformated
            2. if the column type is mime type such as file/png, we store the data of 
            that field in  dictionary of this format {file_path: file_path, binary: binary_data}.
            This is to ensure that we have access to both original data in the dataframe and the 
            binary data in the frontend for this special case'''
        tableData = {}
        tableData['df_id'] = df_id
        tableData['column_names'] = list(df.columns)

        for i, t in enumerate(df.dtypes):
            if t.name not in [CnextMimeType.FILE_PNG, CnextMimeType.FILE_JPG, CnextMimeType.FILE_JPEG, CnextMimeType.URL_PNG, CnextMimeType.URL_JPG, CnextMimeType.URL_JPEG]:
                ## Convert everything else to string #
                df[df.columns[i]] = df[df.columns[i]].apply(str)

        tableData['rows'] = df.values.tolist()

        # Modify data field of column with mime type of file/*. See note above
        #  We chose to do this outside of the dataframe, but it can also be done with a dataframe
        #  see https://stackoverflow.com/questions/41710501/is-there-a-way-to-have-a-dictionary-as-an-entry-of-a-pandas-dataframe-in-python ##
        for i, t in enumerate(df.dtypes):
            if t.name in [CnextMimeType.FILE_PNG, CnextMimeType.FILE_JPG, CnextMimeType.FILE_JPEG]:
                log.info('Load file for mime %s' % t.name)
                for r in range(df.shape[0]):
                    file_path = df[df.columns[i]].iloc[r]
                    log.info('Load file for mime %s path %s' %
                             (t.name, file_path))
                    tableData['rows'][r][i] = {
                        'file_path': file_path,
                        'binary': base64.b64encode(self._get_file_content(file_path, t.name))
                    }
            elif t.name in [CnextMimeType.URL_PNG, CnextMimeType.URL_JPG, CnextMimeType.URL_JPEG]:
                log.info('Load file for mime %s' % t.name)
                for r in range(df.shape[0]):
                    url = df[df.columns[i]].iloc[r]
                    response = requests.get(url)
                    # img = Image.open(BytesIO(response.content))
                    log.info('Load file for mime %s path %s' %
                             (t.name, url))
                    tableData['rows'][r][i] = {
                        'url': url,
                        'binary': base64.b64encode(response.content)
                    }

        tableData['index'] = {}
        tableData['index']['name'] = df.index.name
        tableData['index']['data'] = []
        if re.search(r'datetime', str(df.index.dtype)):
            [tableData['index']['data'].append(str(idx)) for idx in df.index]
        else:
            tableData['index']['data'] = df.index.tolist()
        # log.info(tableData)
        return tableData

    @ipython_internal_output
    def _get_table_data(self, df_id, code):
        output = None
        result = self.user_space.execute(code, ExecutionMode.EVAL)
        # print("get table data %s" % result)
        # log.info("get table data %s" % result)
        if result is not None:
            # log.info("get table data %s" % result)
            output = self._create_table_data(df_id, result)
        return output

    @ipython_internal_output
    def _get_metadata(self, df_id):
        shape = self.user_space.execute("%s.shape" % df_id, ExecutionMode.EVAL)
        dtypes = self.user_space.execute(
            "%s.dtypes" % df_id, ExecutionMode.EVAL)
        countna = self.user_space.execute(
            "%s.isna().sum()" % df_id, ExecutionMode.EVAL)
        describe = self.user_space.execute(
            "%s.describe(include='all')" % df_id, ExecutionMode.EVAL)
        describe = self._convert_to_str_if_not_jsonable(describe)

        columns = {}
        MAX_UNIQUE_LENGTH = 1000
        for col_name, ctype in dtypes.items():
            if re.search(r'datetime', ctype.name):
                ## the unique value of datetime is usually the same as the length so it is meaningless
                ## also we have to convert it to string before sending back. So just don't do it now.
                unique = []                
            else:
                unique = self.user_space.execute(
                    "_pd.Series(%s['%s'].unique()).tolist()" % (df_id, col_name), ExecutionMode.EVAL)
            
            # only send unique list that has length smaller than MAX_UNIQUE_LENGTH
            if len(unique) > MAX_UNIQUE_LENGTH:
                unique = []
            
            columns[col_name] = {'name': col_name,
                                 'type': str(ctype.name), 'unique': unique, 
                                 'countna': countna[col_name].item(), 'describe': describe[col_name].to_dict()}
        output = {'df_id': df_id, 'shape': shape, 'columns': columns}
        return output

    def _process_error_message(self, ipython_message, client_message):
        # log.error("Error %s" % (msg_ipython.content['traceback']))
        if isinstance(ipython_message.content['traceback'], list):
            content = '\n'.join(ipython_message.content['traceback'])
        else:
            content = ipython_message.content['traceback']
        return self._create_error_message(
            WebappEndpoint.DFManager, content, client_message.command_name, client_message.metadata)

    MAX_PLOTLY_SIZE = 1024*1024  # 1MB

    def _create_return_message(self, ipython_message, stream_type, client_message):
        ipython_message = IpythonResultMessage(**ipython_message)
        message = None
        if self._is_error_message(ipython_message.header):
            message = self._process_error_message(
                ipython_message, client_message)
        else:
            result = self.get_execute_result(ipython_message)            
            if result is not None:           
                message = Message(**{'webapp_endpoint': WebappEndpoint.DFManager,
                                    'command_name': client_message.command_name})
                # Add header message from ipython to message metadata
                if message.metadata == None:
                    message.metadata = {}
                message.metadata.update(client_message.metadata)
                message.metadata.update(dict((k, ipython_message.header[k])
                                            for k in ('msg_id', 'msg_type', 'session')))
                message.metadata.update({'stream_type': stream_type})

                if client_message.command_name == DFManagerCommand.plot_column_histogram:
                    if total_size(result) < MessageHandler.MAX_PLOTLY_SIZE:
                        message.content = result
                        message.type = ContentType.RICH_OUTPUT
                        message.sub_type = SubContentType.APPLICATION_PLOTLY
                    else:
                        message.content = "Warning: column histogram plot size is bigger than 1MB -> discard it"
                        message.type = ContentType.STRING
                        message.sub_type = SubContentType.NONE

                elif client_message.command_name == DFManagerCommand.plot_column_quantile:
                    message.content = result
                    message.type = ContentType.RICH_OUTPUT
                    message.sub_type = SubContentType.APPLICATION_PLOTLY

                elif client_message.command_name == DFManagerCommand.get_table_data:
                    # log.info('DFManagerCommand.get_table_data: %s' % result)
                    log.info('DFManagerCommand.get_table_data')
                    message.content = result
                    message.type = ContentType.PANDAS_DATAFRAME
                    message.sub_type = SubContentType.NONE

                elif client_message.command_name == DFManagerCommand.get_df_metadata:
                    log.info('DFManagerCommand.get_df_metadata: %s' % result)
                    message.content = result
                    message.type = ContentType.DICT
                    message.sub_type = SubContentType.NONE
                
                message.error = False
            
        return message

    def message_handler_callback(self, ipython_message, stream_type, client_message):
        try:
            # if self.request_metadata is not None:
            log.info('message_handler_callback: %s' % ipython_message)
            message = self._create_return_message(
                ipython_message=ipython_message, stream_type=stream_type, client_message=client_message)
            if message != None:
                log.info('Message: %s' % message.command_name)
                self._send_to_node(message)
        except:
            trace = traceback.format_exc()
            log.info("Exception %s" % (trace))
            error_message = BaseMessageHandler._create_error_message(
                client_message.webapp_endpoint, trace, client_message.command_name, {})
            self._send_to_node(error_message)

    def handle_message(self, message):
        # send_reply = False
        # message execution_mode will always be `eval` for this sender
        log.info('Got message %s' % message)
        # log.info('Globals: %s' % client_globals)

        ## this step is important to make sure the query work properly in the backend #
        if (isinstance(message.content, str)):
            message.content = message.content.replace("'", '"')

        try:
            if message.command_name == DFManagerCommand.plot_column_histogram:
                # result = eval(message.content, client_globals)
                self.user_space.execute(
                    message.content, ExecutionMode.EVAL, self.message_handler_callback, message)

            elif message.command_name == DFManagerCommand.plot_column_quantile:
                self.user_space.execute(
                    message.content, ExecutionMode.EVAL, self.message_handler_callback, message)

            elif message.command_name == DFManagerCommand.get_table_data:
                # TODO: turn _df_manager to variable
                self.user_space.execute("{}._get_table_data('{}', '{}')".format(
                    IPythonInteral.DF_MANAGER.value, message.metadata['df_id'], message.content), ExecutionMode.EVAL, self.message_handler_callback, message)

            elif message.command_name == DFManagerCommand.get_df_metadata:
                self.user_space.execute("{}._get_metadata('{}')".format(
                    IPythonInteral.DF_MANAGER.value, message.metadata['df_id']), ExecutionMode.EVAL, self.message_handler_callback, message)

            elif message.command_name == DFManagerCommand.reload_df_status:
                active_df_status = self.user_space.get_active_dfs_status()
                active_df_status_message = Message(**{"webapp_endpoint": WebappEndpoint.DFManager, "command_name": message.command_name,
                                                      "seq_number": 1, "type": "dict", "content": active_df_status, "error": False})
                self._send_to_node(active_df_status_message)

        except:
            trace = traceback.format_exc()
            log.error("%s" % (trace))
            error_message = MessageHandler._create_error_message(
                message.webapp_endpoint, trace, message.command_name)
            self._send_to_node(error_message)



import plotly.io as pio
import plotly.express as px
from libs.message import CommandType, ExperimentManagerCommand
from libs.message_handler import BaseMessageHandler
import plotly.graph_objects as go
from mlflow.tracking.client import MlflowClient
import mlflow.tensorflow
import mlflow
import os
import re
import traceback

import pandas
from libs import logs
from libs.message import Message, WebappEndpoint

log = logs.get_logger(__name__)
pio.renderers.default = "json"


def get_checkpoints(mlflowClient: MlflowClient, run_id):
    """Get the list of checkpoints from mlFlow repos

    Args:
        mlflowClient (MlflowClient): [description]
        run_id ([type]): [description]
    Returns:
        if there is a checkoint then return a dictionary of {index: checkpoint_path}, 
        where index is parsed from the number right after the string'cnext_' in the checkpoint name. 
        This index is associated with index of the metrics returned by MlflowClient.get_metric_history.
        else return None
    """
    checkpoints = None
    artifacts = mlflowClient.list_artifacts(run_id, 'checkpoints')
    if (len(artifacts) > 0):
        checkpoints = {}
        for ckpt in artifacts:
            res = re.search(r'(?<=cnext_)\d+', ckpt.path)
            if res:
                index = res.group(0)
                checkpoints[index] = ckpt.path
    return checkpoints


class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue,  user_space=None):
        super(MessageHandler, self).__init__(p2n_queue, user_space)

    def _add_cnext_metadata(self, object, metadata):
        if type(object) == dict:
            object['_cnext_metadata'] = metadata
        elif type(object) == mlflow.entities.run_info.RunInfo:
            object._cnext_metadata = metadata
        else:
            raise TypeError(
                'object type has to be dict or object, got %s' % type(object))
        return object

    RUN_NAME_SHORTEN_LENGTH = 10

    def _get_cnext_run_name(self, mlFlowClient, run_id):
        run = mlFlowClient.get_run(run_id)
        if 'mlflow.runName' in run.data.tags:
            # add underscore prefix to make it consistent with other fileds on mlflow
            return run.data.tags['mlflow.runName']
        else:
            return run_id[:MessageHandler.RUN_NAME_SHORTEN_LENGTH]

    def _get_metric_plots(self, mlflowClient, run_ids):
        metrics_data = {}
        metrics_index = {}
        # run_ids = message.content['run_ids']
        for run_id in run_ids:
            run = mlflowClient.get_run(run_id)
            metric_keys = run.data.metrics.keys()
            for metric in metric_keys:
                metric_history = mlflowClient.get_metric_history(
                    run_id, metric)
                if metric not in metrics_data:
                    metrics_data[metric] = {}
                cnext_run_name = self._get_cnext_run_name(mlflowClient, run_id)
                ## use run cnext name format for column name #
                metrics_data[metric][cnext_run_name] = [
                    m.value for m in metric_history]
                metrics_index = {
                    'step': [m.step for m in metric_history],
                    'timestamp': [m.timestamp for m in metric_history]
                }
        result = {'plots': {}}
        for metric in metrics_data.keys():
            metrics_df = pandas.DataFrame(
                dict([(k, pandas.Series(v))
                      for k, v in metrics_data[metric].items()]),
                index=metrics_index['step']
            )
            # fig = px.line(metrics_df, markers=True)
            fig = go.Figure()
            for col in metrics_df.columns:
                fig.add_trace(go.Scatter(x=metrics_df.index,
                                         y=metrics_df[col], mode='markers+lines', name=col))
                # fig.add_trace(go.Scatter(x=metrics_df.index, y=metrics_df[col], mode='lines', name=col))
            fig.update_layout(
                xaxis_title="steps",
                yaxis_title=metric,
                legend_title="runs",
                font=dict(size=11)
            )
            result['plots'][metric] = fig.to_json()

        # Checkpoins are keyed using cnext run name. It would be better if we key it by the run id instead
        #  but since the trace name in the plot using name not id, we have to use name to make it consistent #
        checkpoints = {}
        name_to_run_ids = {}
        for run_id in run_ids:
            cnext_run_name = self._get_cnext_run_name(mlflowClient, run_id)
            checkpoints[cnext_run_name] = get_checkpoints(mlflowClient, run_id)
            name_to_run_ids[cnext_run_name] = run_id
        log.info(checkpoints)
        if len(checkpoints) > 0:
            self._add_cnext_metadata(
                result, {'checkpoints': checkpoints, 'name_to_run_ids': name_to_run_ids})

        log.info(result)
        return result

    def _get_list_experiments(self, mlflowClient, params):
        running_exp_id = None
        experiments = getattr(
            mlflowClient, ExperimentManagerCommand.list_experiments)(**params)

        ##
        # Identify the running experiment.
        # Since the API does not provide this functionality, use running run as the proxy for running experiments
        # A run is running if it's end_time is None. This might not work when there are multiple running runs
        # initiated from different context
        # #
        for exp in experiments:
            run_infos = mlflowClient.list_run_infos(exp.experiment_id)
            log.info(run_infos)
            if len(run_infos) > 0 and run_infos[0].status == 'RUNNING':
                running_exp_id = exp.experiment_id
                break
        return {'experiments': experiments, 'running_exp_id': running_exp_id}

    def _get_list_run_infos(self, mlflowClient, params):
        active_run_id = None
        run_infos = getattr(
            mlflowClient, ExperimentManagerCommand.list_run_infos)(**params)
        ##
        # Identify the running experiment.
        # Since the API does not provide this functionality, use running run as the proxy for running experiments
        # A run is running if it's end_time is None. This might not work when there are multiple running runs
        # initiated from different context
        # Note: cann't use mlflow.active_run().info.run_uuid here because this is a different python context with
        # that initiated the run. We can improve this by attach the active run id to the message from client.
        # #
        for run_info in run_infos:
            log.info(run_info)
            if run_info.status == 'RUNNING':
                active_run_id = run_info.run_id
                break

        # define run name according to the cnext rule
        for run_info in run_infos:
            self._add_cnext_metadata(
                run_info, {'run_name': self._get_cnext_run_name(mlflowClient, run_info.run_id)})
            # run_info._cnext_metadata = {'run_name': get_run_name(mlFlowClient, run_info.run_id)}
        return {'runs': run_infos, 'active_run_id': active_run_id}

    def _load_artifact_to_local(self, mlflowClient, params):
        local_dir = params['local_dir']
        run_id = params['run_id']
        artifact_path = params['artifact_path']
        if not os.path.exists(local_dir):
            os.mkdir(local_dir)
        local_path = mlflowClient.download_artifacts(
            run_id, artifact_path, local_dir)
        return {'local_path': local_path}

    def _get_runs_data(self, mlflow, params):
        experiment_id = params['experiment_id']
        filter_string = ''
        if 'filter_string' in params:
            filter_string = params['filter_string']
        runs = mlflow.search_runs([experiment_id], filter_string)
            
        run_ids = None
        if 'run_ids' in params:
            run_ids = params['run_ids']
        
        

        pass

    def _create_return_message(self, command_name, content, error=False):
        message = Message(**{'webapp_endpoint': WebappEndpoint.ExperimentManager,
                             'command_name': command_name})
        message.content = content
        message.error = error
        return message

    def handle_message(self, message):
        log.info('Handle ExperimentManager message: %s' % message)
        try:
            params = message.content
            result = None
            if message.type == CommandType.MLFLOW_CLIENT:
                mlflowClient: MlflowClient = mlflow.tracking.MlflowClient(
                    params['tracking_uri'])
                params.pop('tracking_uri')
                if message.command_name == ExperimentManagerCommand.list_experiments:
                    result = self._get_list_experiments(mlflowClient, params)
                elif message.command_name == ExperimentManagerCommand.list_run_infos:
                    result = self._get_list_run_infos(mlflowClient, params)                
                else:
                    result = getattr(
                        mlflowClient, message.command_name)(**params)
            elif message.type == CommandType.MFLOW:
                tracking_uri = params['tracking_uri']
                params.pop('tracking_uri')
                params['uri'] = tracking_uri
                result = getattr(
                    mlflow, message.command_name)(**params)
            elif message.type == CommandType.MLFLOW_OTHERS:
                mlflowClient: MlflowClient = mlflow.tracking.MlflowClient(
                    params['tracking_uri'])
                if message.command_name == ExperimentManagerCommand.get_metric_plots:
                    result = self._get_metric_plots(
                        mlflowClient, message.content['run_ids'])
                elif message.command_name == ExperimentManagerCommand.load_artifacts_to_local:
                    result = self._load_artifact_to_local(mlflowClient, params)
                elif message.command_name == ExperimentManagerCommand.get_runs_data:
                    result = self._get_runs_data(mlflowClient, params)
                
            if result:
                returned_message = self._create_return_message(
                    message.command_name, result)
                self._send_to_node(returned_message)

        except:
            trace = traceback.format_exc()
            log.error("%s" % (trace))
            error_message = MessageHandler._create_error_message(
                message.webapp_endpoint, trace, message.command_name)
            self._send_to_node(error_message)



import os
import simplejson as json

## path to the server config file #
SERVER_CONFIG_PATH = 'server.yaml'
## path to the workspace config file #
WORKSPACE_METADATA_PATH = 'workspace.yaml'
## name of the folder where cnext stores project related information #
CNEXT_PROJECT_FOLDER = '.cnext'
## name of the file where cnext stores project related config #
CNEXT_PROJECT_METADATA_FILE = 'cnext.yaml'
## name of the setting file#
SETTINGS_FILE = 'config.json'


class JsonSerializable:
    def __init__(self, obj):
        self.obj = obj

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, ignore_nan=True)

    def __repr__(self) -> str:
        return self.toJSON()


class DirMetatdata:
    # def __init__(self, **entries):
    #     self.path = None
    #     self.name = None
    #     self.is_file = None
    #     self.__dict__.update(entries)

    def __init__(self, path, name, is_file, deletable, timestamp):
        self.path = path
        self.name = name
        self.is_file = is_file
        self.deletable = deletable
        self.timestamp = timestamp

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, ignore_nan=True)

    def __repr__(self) -> str:
        return self.toJSON()


class FileMetadata(JsonSerializable):
    def __init__(self, **entries):
        self.path = None
        self.name = None
        # self.type = None
        self.executor = None
        self.timestamp = None
        self.__dict__.update(entries)

    # def __init__(self, path, name, timestamp=None, executor=False):
    #     self.path = path
    #     self.name = name
    #     # self.type = None
    #     self.executor = executor

    #     if timestamp != None:
    #         self.timestamp = timestamp

    # def toJSON(self):
    #     return json.dumps(self, default=lambda o: o.__dict__, ignore_nan=True)

    # def __repr__(self) -> str:
    #     return self.toJSON()


class FileContent:
    def __init__(self, content, code_lines=None, timestamp=None):
        # self.__dict__.update(locals())
        self.content = content
        self.code_lines = code_lines
        self.timestamp = timestamp

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, ignore_nan=True)

    def __repr__(self) -> str:
        return self.toJSON()


class ProjectMetadata(JsonSerializable):
    def __init__(self, config):
        self.open_files = []
        self.open_order = []
        if 'open_files' in config and isinstance(config['open_files'], list):
            for file in config['open_files']:
                self.open_files.append(FileMetadata(**file))
        if 'open_order' in config and isinstance(config['open_order'], list) and len(config['open_order']) == len(self.open_files):
            self.open_order = config['open_order']
        else:
            self.open_order = [file.path for file in self.open_files]


class ProjectInfoInWorkspace(JsonSerializable):
    def __init__(self, **entries):
        self.path = None
        self.name = None
        self.id = None
        self.__dict__.update(entries)
        # self.data_path = os.path.join(self.path, CNEXT_PROJECT_FOLDER)
        # self.config_path = os.path.join(
        #     self.data_path, CNEXT_PROJECT_METADATA_FILE)


class WorkspaceMetadata(JsonSerializable):
    """ Infomation about the projects added to this workspace
    """

    def __init__(self, config: dict):
        self.active_project = None
        self.open_projects = []
        projects_config = config
        if 'active_project' in projects_config:
            self.active_project = projects_config['active_project']
        if 'open_projects' in projects_config and isinstance(projects_config['open_projects'], list):
            for project in projects_config['open_projects']:
                self.open_projects.append(ProjectInfoInWorkspace(**project))


class FileManagerMessageParams:
    def __init__(self, params: dict):
        self.norm_path = None
        self.norm_project_path = None
        self.open_order = []
        self.timestamp = None
        self.path = None
        self.project_path = None
        self.is_file = None

        if not isinstance(params, dict):
            return

        if 'path' in params.keys():
            self.path = params['path']
            # avoid creating `./` when the path is empty
            if params['path'] == "":
                self.norm_path = params['path']
            else:
                self.norm_path = os.path.normpath(params['path'])
        if 'project_path' in params.keys():
            self.project_path = params['project_path']
            self.norm_project_path = os.path.normpath(params['project_path'])
        if 'open_order' in params.keys() and isinstance(params['open_order'], list):
            self.open_order = params['open_order']
        if 'timestamp' in params.keys():
            self.timestamp = params['timestamp']
        if 'is_file' in params.keys():
            self.is_file = params['is_file']



import simplejson as json


class JsonSerializable:
    def __init__(self, obj):
        self.obj = obj

    def toJSON(self):
        return json.dumps(self.obj, ignore_nan=True)

    def __repr__(self) -> str:
        return self.toJSON()


def ipython_internal_output(func):
    '''
    Wrapper to return JsonSerializable instead of original object when return result inside ipython
    '''
    def json_serializable_output(*args, **kwargs):
        output = func(*args, **kwargs)
        return JsonSerializable(output)
    return json_serializable_output



import os
from os.path import isfile, join, exists
import traceback
import send2trash
import simplejson as json
from libs import logs
from project_manager.interfaces import FileMetadata, DirMetatdata
from project_manager.interfaces import FileContent

log = logs.get_logger(__name__)

# The folder that consist the state, configuration files
CNEXT_FOLDER_PATH = '.cnext'

UNDELETABLE_FILES = ['main.py', 'config.json']


def list_dir(project_path, relative_dir_path):
    dir_list: DirMetatdata = []
    try:
        if project_path != None and relative_dir_path != None:
            dir_path = os.path.join(project_path, relative_dir_path)
            log.info('list_dir paths: %s %s %s' %
                     (project_path, relative_dir_path, dir_path))
            for file_name in os.listdir(dir_path):
                relative_file_path = os.path.join(relative_dir_path, file_name)
                file_path = os.path.join(dir_path, file_name)
                new_file = DirMetatdata(
                    relative_file_path,
                    name=file_name,
                    is_file=isfile(file_path),
                    deletable=False if relative_file_path in UNDELETABLE_FILES else True,
                    timestamp=os.path.getmtime(file_path)
                )
                dir_list.append(new_file)
        return dir_list
    except Exception as error:
        log.error("%s - %s" % (error, traceback.format_exc()))
        raise Exception


def get_file_metadata(path):
    try:
        pass
    except Exception:
        raise Exception


def set_name(path):
    try:
        pass
    except Exception:
        raise Exception


def read_file(project_path, relative_file_path, timestamp=None):
    try:
        result: FileContent = None
        if project_path != None and relative_file_path != None:
            log.info('Read file %s, old timestamp %s' %
                     (relative_file_path, timestamp))
            file_path = get_abs_file_path(project_path, relative_file_path)
            if (timestamp != os.path.getmtime(file_path)):
                timestamp = os.path.getmtime(file_path)
                log.info('Read file have new timestamp %s, %s' %
                         (file_path, timestamp))
                # Read file content
                with open(file_path) as file:
                    content = file.read()
                    content_lines = content.splitlines()
                    if content_lines == [] or content[-1] == '\n':
                        content_lines.append('')
                    result = FileContent(
                        content=content_lines, timestamp=timestamp)

                # If state file path is exists that mean, the state already saved.
                # Read data from state file path.
                state_path = get_state_path(project_path, relative_file_path)
                if os.path.exists(state_path):
                    with open(state_path) as state_file:
                        data = json.load(state_file)
                        result.code_lines = data
            else:
                log.info('Read file have the same timestamp %s, %s' %
                         (relative_file_path, timestamp))
        return result
    except Exception:
        raise Exception


def create_file(project_path, relative_file_path):
    try:
        if project_path != None and relative_file_path != None:
            log.info('Create file %s' % relative_file_path)
            file_path = get_abs_file_path(project_path, relative_file_path)
            if exists(file_path):  # TODO return error here
                log.info('File already exists %s' % file_path)
            else:
                with open(file_path, 'w') as file:
                    pass
            return FileMetadata(path=relative_file_path, name=relative_file_path.split('/')[-1], timestamp=os.path.getmtime(file_path))
        return None
    except Exception as error:
        log.error("%s - %s" % (error, traceback.format_exc()))
        raise Exception


def create_folder(project_path, relative_folder_path):
    try:
        if project_path != None and relative_folder_path != None:
            log.info('Create folder %s' % relative_folder_path)
            folder_path = get_abs_file_path(project_path, relative_folder_path)
            if exists(folder_path):  # TODO return error here
                log.info('Folder already exists %s' % folder_path)
            else:
                os.mkdir(folder_path)
            return FileMetadata(path=relative_folder_path, name=relative_folder_path, timestamp=os.path.getmtime(folder_path))
        return None
    except Exception as error:
        log.error("%s - %s" % (error, traceback.format_exc()))
        raise Exception


def delete(project_path, relative_file_path):
    try:
        if project_path != None and relative_file_path != None:
            file_path = get_abs_file_path(project_path, relative_file_path)
            log.info('Send item to trash %s' % file_path)
            send2trash.send2trash(file_path)
            return FileMetadata(path=project_path, name=relative_file_path, timestamp=None)
        return None
    except Exception as error:
        log.error("%s - %s" % (error, traceback.format_exc()))
        raise Exception


def save_file(project_path, relative_file_path, content):
    """
        Save file and return file metadata
    """
    try:
        if project_path != None and relative_file_path != None:
            log.info('Save file {}'.format(relative_file_path))
            file_path = get_abs_file_path(project_path, relative_file_path)
            file_name = relative_file_path.split('/')[-1]
            with open(file_path, 'w') as file:
                file.write(content)
            return FileMetadata(path=relative_file_path, name=file_name, timestamp=os.path.getmtime(file_path))
        else:
            return None
    except Exception as error:
        log.error("%s - %s" % (error, traceback.format_exc()))
        raise Exception


def save_state(project_path, relative_file_path, content):
    """
        Save state and return file metadata
    """
    try:
        if project_path != None and relative_file_path != None:
            log.info('Save state {}'.format(relative_file_path))
            state_file_path = get_state_path(project_path, relative_file_path)
            state_file_name = state_file_path.split('/')[-1]
            os.makedirs(os.path.dirname(state_file_path), exist_ok=True)
            with open(state_file_path, 'w') as outfile:
                outfile.write(json.dumps(content))
            # TODO return the relative state path instead
            return FileMetadata(path=relative_file_path, name=state_file_name, timestamp=os.path.getmtime(state_file_path))
        else: 
            return None
    except Exception as ex:
        log.error("Save state error: {}".format(ex))
        raise Exception


def get_abs_file_path(project_path, relative_file_path):
    return os.path.join(project_path, relative_file_path)


def get_state_path(project_path, relative_file_path):
    """
        Get state path from file path
    """
    # if project_path not in relative_file_path:
    #     err = "Project path is not match with file path: {project_path} path: {path}".format(
    #         project_path=project_path, path=relative_file_path)
    #     log.error(err)
    #     raise Exception(err)
    # sub_path = relative_file_path.replace(project_path, "")
    # # If the first character is slash, must remove this to make the os.path.join function work properly
    # # If It start with a slash, then python is considered an "absolute path" and everything before it is discarded
    # if sub_path[0] == '/':
    #     sub_path = sub_path[1::]  # Remove the first character is slash
    file_path = os.path.join(
        project_path, CNEXT_FOLDER_PATH, 'states', relative_file_path)
    state_path = os.path.splitext(file_path)[0] + '.json'
    return state_path



import zmq


# from libs import logs
# log = logs.get_logger(__name__)
# class MessageQueue:
#     def __init__(self, host: str, port: int, type=MessageQueueType.PULL, hwm=1000):
#         context = zmq.Context()
#         self.host = host
#         self.port = port
#         self.addr = '{}:{}'.format(host, port)
#         if type == MessageQueueType.PUSH:
#             self.push: zmq.Socket = context.socket(zmq.PUSH)
#             self.push.connect(self.addr)
#         elif type == MessageQueueType.PULL:
#             self.pull: zmq.Socket = context.socket(zmq.PULL)
#             self.pull.connect(self.addr)

#     def get_socket(self):
#         return self.push

#     def send(self, message):
#         # TODO: could not explain why NOBLOCK would not work even when the receiver already connects
#         # TODO: and calls `receive` command
#         # res = self.push_queue.send_json(message)#, flags=zmq.NOBLOCK)
#         res = self.push.send_string(message)
#         return res

#     def close(self):
#         self.pull_queue.disconnect(self.addr)

#     def receive_msg(self):
#         return self.pull.recv()


from libs.config import read_config
from project_manager.interfaces import SERVER_CONFIG_PATH
config = read_config(SERVER_CONFIG_PATH, {'code_executor_comm': {
    'host': '127.0.0.1', 'n2p_port': 5001, 'p2n_port': 5002}})


class MessageQueuePush:
    def __init__(self, host, port, hwm=1000):
        context = zmq.Context()
        self.host = host
        self.port = port
        self.addr = '{}:{}'.format(self.host, self.port)
        self.push: zmq.Socket = context.socket(zmq.PUSH)
        self.push.connect(self.addr)

    def get_socket(self):
        return self.push

    def send(self, message):
        # TODO: could not explain why NOBLOCK would not work even when the receiver already connects
        # TODO: and calls `receive` command
        # res = self.push_queue.send_json(message)#, flags=zmq.NOBLOCK)
        res = self.push.send_string(message)
        return res


class MessageQueuePull:
    def __init__(self, host, port):
        self.context = zmq.Context()
        self.context.setsockopt(zmq.LINGER, 0)
        self.host = host
        self.port = port
        self.addr = '{}:{}'.format(self.host, self.port)
        self.pull: zmq.Socket = self.context.socket(zmq.PULL)
        self.pull.bind(self.addr)

    def close(self):
        self.pull_queue.disconnect(self.addr)

    def receive_msg(self):
        return self.pull.recv_string()



from enum import Enum

from libs.json_serializable import JsonSerializable


class IpythonResultMessage(JsonSerializable):
    def __init__(self, **entries):
        self.header = None
        self.msg_id = None
        self.msg_type = None
        self.parent_header = None
        self.metadata = None
        self.content = None
        self.buffers = None
        self.__dict__.update(entries)


class IPythonConstants:
    class MessageType(str, Enum):
        EXECUTE_REPLY = 'execute_reply'
        INSPECT_REPLY = 'inspect_reply'
        COMPLETE_REPLY = 'complete_reply'
        HISTORY_REPLY = 'history_reply'
        IS_COMPLETE_REPLY = 'is_complete_reply'
        CONNECT_REPLY = 'connect_reply'
        COMM_INFO_REPLY = 'comm_info_reply'
        KERNEL_INFO_REPLY = 'kernel_info_reply'
        SHUTDOWN_REPLY = 'shutdown_reply'
        INTERRUPT_REPLY = 'interrupt_reply'
        DEBUG_REPLY = 'debug_reply'
        DISPLAY_DATA = 'display_data'
        UPDATE_DISPLAY_DATA = 'update_display_data'
        EXECUTE_INPUT = 'execute_input'
        EXECUTE_RESULT = 'execute_result'
        STREAM = 'stream'
        ERROR = 'error'
        STATUS = 'status'
        CLEAR_OUTPUT = 'clear_output'
        DEBUG_EVENT = 'debug_event'
        INPUT_REPLY = 'input_reply'

    class StreamType(str, Enum):
        IOBUF = 'iobuf'
        SHELL = 'shell'
        CONTROL = 'control'

    class ShellMessageStatus(str, Enum):
        OK = 'ok'
        ERROR = 'error'
        IDLE = 'idle'

    class IOBufMessageStatus(str, Enum):
        OK = 'ok'
        ERROR = 'error'
        ABORT = 'abort'

    class ExecutionState(str, Enum):
        STARTING = 'starting'
        BUSY = 'busy'
        IDLE = 'idle'


class IPythonInteral(Enum):
    DF_MANAGER = '_df_manager'
    CASSIST = '_cassist'
    USER_SPACE = '_user_space'



import cnextlib.dataframe as cd
import plotly.express as px
import plotly.io as pio
import io
import sys
import os
os.chdir('/Volumes/GoogleDrive/.shortcut-targets-by-id/1FrvaCWSo3NV1g0sR9ib6frv_lzRww_8K/CycAI/works/CAT/machine_simulation/')
sys.path.append(os.getcwd())
training_data = cd.DataFrame('data/exp_data/997/21549286_out.csv')
df = training_data


def test_a_simple_line_graph():
    # pio.renderers.default = "json"
    fig = px.line(df, x=df.index, y='Fuel Rail Pressure', title="Machine")
    fig.show()


def px_original_test():
    # pio.renderers.default = "json"
    long_df = px.data.medals_long()
    fig = px.bar(long_df, x="nation", y="count",
                 color="medal", title="Long-Form Input")
    normal_stdout = sys.stdout
    sys.stdout = io.StringIO()
    fig.show()
    sys.stdout = normal_stdout


def test_histogram():
    # pio.renderers.default = "json"
    fig = eval("px.histogram(df, x='Fuel Rail Pressure')")
    print(type(fig))
    # fig.show()

# test_histogram()



from enum import Enum
import threading
import traceback
import simplejson as json
import cnextlib.user_space as _cus
import cnextlib.dataframe as _cd
from libs.constants import TrackingModelType
from user_space.ipython.kernel import IPythonKernel
from user_space.ipython.constants import IPythonInteral, IPythonConstants

from libs import logs
log = logs.get_logger(__name__)


class ExecutionMode(Enum):
    EVAL = 0
    EXEC = 1


class BaseKernel:
    def __init__(self) -> None:
        pass

    def _assign_exec_mode(self, code):
        exec_mode = 'eval'
        try:
            compile(code, '<stdin>', 'eval')
        except SyntaxError as error:
            log.error(error)
            exec_mode = 'exec'

        log.info("assigned command type: %s" % exec_mode)
        return exec_mode

    def execute(self, code, exec_mode: ExecutionMode = None, userspace_globals=globals()):
        if exec_mode == None:
            exec_mode = self._assign_exec_mode(code)
        if exec_mode == ExecutionMode.EVAL:
            return eval(code, userspace_globals)
        elif exec_mode == ExecutionMode.EXEC:
            return exec(code, userspace_globals)


class IPythonUserSpace(_cus.UserSpace):
    ''' 
        Define the space where user code will be executed. 
        This is encapsulated in a python module so all the imports and variables are separated out from the rest.
        The code is executed on a kernel such as BaseKernel or IPythonKernel
    '''

    def __init__(self, tracking_df_types: tuple = (), tracking_model_types: tuple = ()):
        self.executor: IPythonKernel = IPythonKernel()
        self.init_executor()
        self.execution_lock = threading.Lock()
        self.result = None
        super().__init__(tracking_df_types, tracking_model_types)

    def init_executor(self):
        code = """
import cnextlib.dataframe as _cd
import pandas as _pd
from dataframe_manager import dataframe_manager as _dm
from cassist import cassist as _ca
from user_space.user_space import BaseKernelUserSpace

## need to create a new _UserSpace class here so that the global() will be represent this module where all the execution are #
class _UserSpace(BaseKernelUserSpace):
    def globals(self):
        return globals()

{_user_space} = _UserSpace((_cd.DataFrame, _pd.DataFrame), {tracking_models})  
{_df_manager} = _dm.MessageHandler(None, {_user_space})
{_cassist} = _ca.MessageHandler(None, {_user_space})
""".format(_user_space=IPythonInteral.USER_SPACE.value,
           _df_manager=IPythonInteral.DF_MANAGER.value,
           _cassist=IPythonInteral.CASSIST.value,
           tracking_models=(TrackingModelType.PYTORCH_NN,TrackingModelType.TENSORFLOW_KERAS))
        self.executor.execute(code)

    def globals(self):
        return globals()

    def _complete_execution_message(self, message) -> bool:
        return message['header']['msg_type'] == 'execute_reply' and 'status' in message['content']

    def message_handler_callback(self, ipython_message, stream_type, client_message):
        try:
            log.info('%s msg: %s %s' % (
                stream_type, ipython_message['header']['msg_type'], ipython_message['content']))
            # log.info('%s msg: msg_type = %s' % (
            #     stream_type, ipython_message['header']['msg_type']))
            if ipython_message['header']['msg_type'] == IPythonConstants.MessageType.EXECUTE_RESULT:
                self.result = json.loads(
                    ipython_message['content']['data']['text/plain'])
            elif self._complete_execution_message(ipython_message) and self.execution_lock.locked():
                self.execution_lock.release()
                log.info('Execution unlocked')
            else:
                # TODO: log everything else
                log.info('Other messages: %s' % ipython_message)
        except:
            # this is internal exception, we won't send it to the client
            trace = traceback.format_exc()
            log.info("Exception %s" % (trace))

    def _result_waiting_execution(func):
        '''
        Wrapper to block the execution until the execution complete
        '''
        def _result_waiting_execution_wrapper(*args, **kwargs):
            args[0].execution_lock.acquire()
            log.info('User_space execution lock acquired')
            func(*args, **kwargs)
            args[0].execution_lock.acquire()
            args[0].execution_lock.release()
            log.info('User_space execution lock released')
            log.info("Results: %s" % args[0].result)
            return args[0].result
        return _result_waiting_execution_wrapper

    @_result_waiting_execution
    def get_active_dfs_status(self):
        """ This function will be blocked until the execution completes and the result will be returned directly from here """
        """Generate the list of dfs status from execution
        Note: there might be multiple updates happened to a dataframe during multiline execution, 
        therefore the `result` will be a list.

        Returns:
            _type_: _description_
        """
        code = "{_user_space}.get_active_dfs_status()".format(
            _user_space=IPythonInteral.USER_SPACE.value)
        log.info('Code to execute %s' % code)
        self.executor.execute(code, None, self.message_handler_callback)

    @_result_waiting_execution
    def get_active_models_info(self):
        """ This function will be blocked until the execution completes and the result will be returned directly from here """
        code = "{_user_space}.get_active_models_info()".format(
            _user_space=IPythonInteral.USER_SPACE.value)
        log.info('Code to execute %s' % code)
        self.executor.execute(code, None, self.message_handler_callback)

    def reset_active_dfs_status(self):
        code = "{_user_space}.reset_active_dfs_status()".format(
            _user_space=IPythonInteral.USER_SPACE.value)
        self.executor.execute(code)

    def execute(self, code, exec_mode: ExecutionMode = None, message_handler_callback=None, client_message=None):
        self.reset_active_dfs_status()
        return self.executor.execute(code, exec_mode, message_handler_callback, client_message)

    def shutdown_executor(self):
        self.executor.shutdown_kernel()
        if self.execution_lock.locked():
            self.execution_lock.release()

    def restart_executor(self):
        result = self.executor.restart_kernel()
        self.init_executor()
        if self.execution_lock.locked():
            self.execution_lock.release()
        return result

    def interrupt_executor(self):
        result = self.executor.interrupt_kernel()
        if self.execution_lock.locked():
            self.execution_lock.release()
        return result
    
    def set_executor_working_dir(self, path):
        code = "import os; os.chdir('{}')".format(path)
        return self.executor.execute(code)


class BaseKernelUserSpace(_cus.UserSpace):
    ''' 
        Define the space where user code will be executed. 
        This is encapsulated in a python module so all the imports and variables are separated out from the rest.
        The code is executed on a kernel such as BaseKernel or IPythonKernel
    '''

    def __init__(self, tracking_df_types: tuple = (), tracking_model_types: tuple = ()):
        self.executor = BaseKernel()
        ## need to set user space on DataFrameTracker, it does not work if set with DataFrame
        _cd.DataFrameTracker.set_user_space(self)
        super().__init__(tracking_df_types, tracking_model_types)

    def globals(self):
        return globals()

    def execute(self, code, exec_mode: ExecutionMode = None):
        ## this function is not called when using ipython
        # self.reset_active_dfs_status()
        return self.executor.execute(code, exec_mode, self.globals())

    def shutdown_executor(self):
        pass



import traceback, sys

try:
    eval('d = long_df')
except:
    traceback.print_exc()
    print(sys.exc_info())

try:
    eval('long_df_')
except:
    traceback.print_exc()    
    print(sys.exc_info())


import threading
import time
import traceback
import jupyter_client
import queue
from user_space.ipython.constants import IPythonConstants
from libs import logs
log = logs.get_logger(__name__)

MESSSAGE_TIMEOUT = 1

class IPythonKernel():
    def __init__(self):
        self.km = jupyter_client.KernelManager()
        self.km.start_kernel()
        self.kc = self.km.blocking_client()
        self.wait_for_ready()

        self.stop_stream_thread = False
        self.shell_msg_thread = threading.Thread(
            target=self.handle_ipython_stream, args=(IPythonConstants.StreamType.SHELL,), daemon=True)
        self.iobuf_msg_thread = threading.Thread(
            target=self.handle_ipython_stream, args=(IPythonConstants.StreamType.IOBUF,), daemon=True)
        self.shell_msg_thread.start()
        self.iobuf_msg_thread.start()

        self.message_handler_callback = None
        ## This lock is used to make sure only one execution is being executed at any moment in time #
        self.execute_lock = threading.Lock()
        self._set_execution_complete_condition(False)

    def shutdown_kernel(self):
        try:
            if self.km.is_alive():
                log.info('Kernel shutting down')
                self.kc.stop_channels()
                self.km.shutdown_kernel(now=True)
                log.info('Kernel shutdown')
        except:
            trace = traceback.format_exc()
            log.info("Exception %s" % (trace))

    def restart_kernel(self):
        try:
            if self.km.is_alive():
                log.info('Kernel restarting')
                self.km.restart_kernel()
                self.stop_stream_thread = True
                self.kc = self.km.blocking_client()
                self.wait_for_ready()
                # wait to make sure the stream threads will stop before proceeding
                while self.shell_msg_thread.is_alive() or self.shell_msg_thread.is_alive():
                    time.sleep(1)
                self.shell_msg_thread = threading.Thread(
                    target=self.handle_ipython_stream, args=(IPythonConstants.StreamType.SHELL,), daemon=True)
                self.iobuf_msg_thread = threading.Thread(
                    target=self.handle_ipython_stream, args=(IPythonConstants.StreamType.IOBUF,), daemon=True)
                self.stop_stream_thread = False
                self.shell_msg_thread.start()
                self.iobuf_msg_thread.start()
                log.info('Kernel restarted')

                # release execution lock which might be locked during an execution
                if self.execute_lock.locked():
                    self.execute_lock.release()
                    log.info('Kernel execution lock released')
                    self._set_execution_complete_condition(True)
                return True
        except:
            trace = traceback.format_exc()
            log.info("Exception %s" % (trace))
        return False

    def interrupt_kernel(self):
        try:
            if self.km.is_alive():
                self.km.interrupt_kernel()
                log.info('Interupt kernel')
            return True
        except:
            trace = traceback.format_exc()
            log.info("Exception %s" % (trace))
        return False

    def wait_for_ready(self):
        try:
            self.kc.wait_for_ready(timeout=50)
        except RuntimeError:
            self.shutdown_kernel()

    def _is_status_message(self, message):
        return message['header']['msg_type'] == 'status'

    def _set_execution_complete_condition(self, condition: bool):
        self.shell_cond = condition
        self.iobuf_cond = condition

    def _set_execution_complete_condition_from_message(self, stream_type, message):
        if stream_type == IPythonConstants.StreamType.SHELL:
            self.shell_cond = message['header']['msg_type'] == 'execute_reply' and 'status' in message['content']
        if stream_type == IPythonConstants.StreamType.IOBUF:
            self.iobuf_cond = message['header']['msg_type'] == 'status' and message['content']['execution_state'] == 'idle'

    def _is_execution_complete(self, stream_type, message) -> bool:
        ## Look at the shell stream and check the status #
        return self.shell_cond and self.iobuf_cond

    def handle_ipython_stream(self, stream_type: IPythonConstants.StreamType):
        try:
            log.info('Start stream %s' % stream_type)
            while not self.stop_stream_thread:
                try:
                    ipython_message = None
                    if stream_type == IPythonConstants.StreamType.SHELL:
                        ipython_message = self.kc.get_shell_msg(
                            timeout=MESSSAGE_TIMEOUT)
                    elif stream_type == IPythonConstants.StreamType.IOBUF:
                        ipython_message = self.kc.get_iopub_msg(
                            timeout=MESSSAGE_TIMEOUT)

                    if 'status' in ipython_message['content']:
                        log.info('%s msg: msg_type = %s, status = %s' % (
                            stream_type, ipython_message['header']['msg_type'], ipython_message['content']['status']))
                    elif ipython_message['header']['msg_type'] == 'status' or ipython_message['header']['msg_type'] == 'error':
                        log.info('%s msg: msg_type = %s, content = %s' % (
                            stream_type, ipython_message['header']['msg_type'], ipython_message['content']))
                    else:
                        log.info('%s msg: msg_type = %s' % (
                            stream_type, ipython_message['header']['msg_type']))
                        # log.info('%s msg: msg_type = %s msg_content = %s' % (
                        #     stream_type, ipython_message['header']['msg_type'], ipython_message['content']))

                    if ipython_message is not None and self.message_handler_callback is not None:
                        self.message_handler_callback(
                            ipython_message, stream_type, self.client_message)

                    ## unlock execute lock only after upstream has processed the data if messge is status #
                    # if self._is_status_message(ipython_message) and ipython_message['content']['execution_state'] != 'busy' and self.execute_lock.locked():
                    #     self.execute_lock.release()
                    self._set_execution_complete_condition_from_message(
                        stream_type, ipython_message)
                    if self.execute_lock.locked() and self._is_execution_complete(stream_type, ipython_message):
                        self.execute_lock.release()
                        log.info('Kernel execution lock released')
                except queue.Empty:
                    pass
            log.info('Stop stream %s' % stream_type)
        except:
            trace = traceback.format_exc()
            log.info("Exception %s" % (trace))

    def execute(self, code, exec_mode=None, message_handler_callback=None, client_message=None):
        self.execute_lock.acquire()
        log.info('Kernel execution lock acquired')
        self._set_execution_complete_condition(False)
        self.message_handler_callback = message_handler_callback
        self.client_message = client_message
        self.kc.execute(code)



import sys, logging, json, io
import pandas 

logging.basicConfig(filename='./log.txt', filemode='a', format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s', 
                        datefmt='%H:%M:%S', level=logging.DEBUG)
log = logging.getLogger(__name__)

while True:
    for line in sys.stdin:        
        try:
            req = json.loads(line)        
            log.info(req)
            exec(req['command'])
        except:
            import traceback
            log.error(traceback.format_exc())
            # print(json.dumps({"commandType": req['command_type'], "contentType": "str", "content": traceback.format_exc(), "error": True}))                        
            traceback.print_exc();
        sys.stdout.flush()     




import os
import uuid
from os.path import exists
import simplejson as json
import traceback
from project_manager.interfaces import ProjectInfoInWorkspace
from libs import logs
from libs.config import read_config, save_config
from project_manager.interfaces import FileMetadata, CNEXT_PROJECT_METADATA_FILE, CNEXT_PROJECT_FOLDER, WORKSPACE_METADATA_PATH, WorkspaceMetadata, SETTINGS_FILE, ProjectMetadata

log = logs.get_logger(__name__)

active_project = None
CNEXT_PROJECT_DIR = ''

## File #


def get_project_metadata_path(project):
    """ Get path to .cnext/cnext.yaml """
    return os.path.join(project['path'], CNEXT_PROJECT_FOLDER, CNEXT_PROJECT_METADATA_FILE)


def get_open_files():
    """ Get active project's open files """
    open_files = []
    try:
        active_project = get_active_project()
        # config_file_path = active_project['config_path']
        config_file_path = get_project_metadata_path(active_project)
        if exists(config_file_path):
            config = read_config(config_file_path)
            project_metadata = ProjectMetadata(config.__dict__)
        else:
            log.error("Config file does not exist %s" % (config_file_path))
        return project_metadata
    # except yaml.YAMLError as error:
    #     log.error("%s" % (error))
    #     return []
    except Exception as error:
        log.error("%s - %s" % (error, traceback.format_exc()))
        raise Exception  # this will be seen in the web app #


def close_file(path, open_order):
    new_open_files = []
    try:
        if path != None:
            active_project = get_active_project()
            # config_file_path = active_project['config_path']
            config_file_path = get_project_metadata_path(active_project)
            if exists(config_file_path):
                config = read_config(config_file_path)
                project_metadata = ProjectMetadata(config.__dict__)
                for file in project_metadata.open_files:
                    if(file.path != path):
                        new_open_files.append(file)                        
                project_metadata.open_files = new_open_files

                if isinstance(open_order, list):
                    ## update with the latest order #
                    project_metadata.open_order = open_order
                if path in open_order:
                    ## remove the path #
                    project_metadata.open_order.remove(path)

                save_config(project_metadata, config_file_path)
            else:
                log.error("Config file does not exist %s" % (config_file_path))
        return project_metadata
    # except yaml.YAMLError as error:
    #     log.error("%s" % (error))
    #     return []
    except Exception as error:
        log.error("%s - %s" % (error, traceback.format_exc()))
        raise Exception  # this will be seen in the web app #


def open_file(path, open_order):
    try:
        if path != None:
            active_project = get_active_project()
            # config_path = active_project['config_path']
            project_metadata_path = get_project_metadata_path(active_project)
            if exists(project_metadata_path):
                config = read_config(project_metadata_path)
                project_metadata = ProjectMetadata(config.__dict__)
                file_existed = [
                    file_item for file_item in project_metadata.open_files if file_item.path == path]
                if len(file_existed) == 0:
                    ## Note that we dont set the timestamp when open the file #
                    file = FileMetadata(path=path,
                                        name=path.split('/')[-1])
                    project_metadata.open_files.append(file)
                    
                    if isinstance(open_order, list):
                        ## update with the latest order #
                        project_metadata.open_order = open_order
                    if path in project_metadata.open_order:
                        ## remove path before append it to the end #
                        project_metadata.open_order.remove(path)
                    project_metadata.open_order.append(path)
                    
                    # config.open_files = open_files
                    save_config(project_metadata, project_metadata_path)
            else:
                log.error("Config file does not exist %s" %
                          (project_metadata_path))
            return project_metadata
        else:
            return []
    # except yaml.YAMLError as error:
    #     log.error("%s" % (error))
    #     return []
    except Exception as error:
        log.error("%s - %s" % (error, traceback.format_exc()))
        raise Exception  # this will be seen in the web app #
##

## Project #


def set_project_dir(path):
    global CNEXT_PROJECT_DIR
    CNEXT_PROJECT_DIR = path
    return True


def set_active_project(project_id: str):
    config = read_config(WORKSPACE_METADATA_PATH)
    config_dict = config.__dict__
    config_dict['active_project'] = project_id
    return save_workspace_metadata(config=config_dict)


def get_active_project():
    config = read_config(WORKSPACE_METADATA_PATH)
    config_dict = config.__dict__
    active_project = [project for project in config_dict['open_projects']
                      if project['id'] == config_dict['active_project']]
    if len(active_project) == 0:
        log.error("Not found active project!")
        raise Exception
    return active_project[0]


def set_working_dir(path):
    try:
        os.chdir(path)
        return True
    except Exception:
        raise Exception
##

## Project Settings #


def get_project_settings_path(project):
    """ Get path to config.json """
    return os.path.join(project['path'], SETTINGS_FILE)


def save_project_settings(content):
    """ Save config.json"""
    try:
        active_project = get_active_project()
        settings_file_path = get_project_settings_path(active_project)
        os.makedirs(os.path.dirname(settings_file_path), exist_ok=True)
        with open(settings_file_path, 'w') as outfile:
            outfile.write(json.dumps(content, indent=4))
        return FileMetadata(path=settings_file_path, name=SETTINGS_FILE, timestamp=os.path.getmtime(settings_file_path))
    except Exception as ex:
        log.error("Save config error: {}".format(ex))
        raise Exception


def get_project_settings():
    """ Save config from config.json"""
    active_project = get_active_project()
    config_file_path = get_project_settings_path(active_project)
    if os.path.exists(config_file_path):
        config_file_data = open(config_file_path, "r")
        data = json.loads(config_file_data.read())
        return data
    return
##

## Workspace #


def add_project(path):
    try:
        project_name = path.split('/')[-1]

        if not os.path.exists(path):
            os.mkdir(path)

        # Update workspace config
        config = read_config(WORKSPACE_METADATA_PATH)
        workspace_metadata = WorkspaceMetadata(config.__dict__)
        # config_dict = config.__dict__
        exist_project = [
            project for project in workspace_metadata.open_projects if project.path == path]
        if len(exist_project) > 0:
            project_id = exist_project[0].id
        else:
            project_id = str(uuid.uuid1())
            new_project = {
                'id': project_id,
                'name': project_name,
                'path': path
            }
            workspace_metadata.open_projects.append(new_project)
            workspace_metadata.active_project = project_id
            save_config(workspace_metadata, WORKSPACE_METADATA_PATH)

            # # Create main.py file
            # main_file_path = os.path.join(path, 'main.py')
            # if not os.path.exists(main_file_path):
            #     with open(main_file_path, 'w'):
            #         pass

        # Assign project
        active_project = ProjectInfoInWorkspace(
            path=path,
            name=project_name,
            id=project_id,
            # data_path=os.path.join(path, CNEXT_PROJECT_FOLDER),
            # config_path=None
        )

        # active_project.config_path = os.path.join(
        #     active_project.data_path, CNEXT_PROJECT_METADATA_FILE)

        # Set activate project
        # set_active_project(active_project)

        # Create .cnext/cnext.yaml if not exsists
        cnext_project_path = os.path.join(path, CNEXT_PROJECT_FOLDER)
        if not os.path.exists(cnext_project_path):
            os.mkdir(cnext_project_path)
        cnext_config_path = os.path.join(
            cnext_project_path, CNEXT_PROJECT_METADATA_FILE)
        if not os.path.exists(cnext_config_path):
            with open(cnext_config_path, 'w'):
                pass

        return workspace_metadata
    except Exception as ex:
        raise ex


def get_workspace_metadata():
    config = read_config(WORKSPACE_METADATA_PATH)
    return config.__dict__


def save_workspace_metadata(config):
    workspace_info = WorkspaceMetadata(config)
    save_config(workspace_info, WORKSPACE_METADATA_PATH)
    return workspace_info
##



from libs.json_serializable import JsonSerializable

class ModelInfo:
    def __init__(self, **entries):
        self.name = None
        self.id = None
        self.obj_class = None
        self.base_class = None
        self.__dict__.update(entries)

class NetronStatus:
    OK = 'ok'
    ERROR = 'error'


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cyc_next_app',
 'cyc_next_app.lib.@jupyter-widgets.schema',
 'cyc_next_app.server',
 'cyc_next_app.server.python',
 'cyc_next_app.server.python.cassist',
 'cyc_next_app.server.python.code_editor',
 'cyc_next_app.server.python.dataframe_manager',
 'cyc_next_app.server.python.experiment_manager',
 'cyc_next_app.server.python.file_explorer',
 'cyc_next_app.server.python.file_manager',
 'cyc_next_app.server.python.libs',
 'cyc_next_app.server.python.model_manager',
 'cyc_next_app.server.python.project_manager',
 'cyc_next_app.server.python.user_space',
 'cyc_next_app.server.python.user_space.ipython',
 'cyc_next_app.server.tests',
 'cyc_next_app.server.tests.test_servers']

package_data = \
{'': ['*'],
 'cyc_next_app': ['lib/@jupyter-widgets/base/*',
                  'lib/@jupyter-widgets/base/css/*',
                  'lib/@jupyter-widgets/base/lib/*',
                  'lib/@jupyter-widgets/controls/*',
                  'lib/@jupyter-widgets/controls/css/*',
                  'lib/@jupyter-widgets/controls/lib/*',
                  'lib/@jupyter-widgets/controls/lib/phosphor/*',
                  'lib/@jupyter-widgets/html-manager/*',
                  'lib/@jupyter-widgets/html-manager/css/*',
                  'lib/@jupyter-widgets/html-manager/lib/*',
                  'lib/@jupyter-widgets/jupyterlab-manager/*',
                  'lib/@jupyter-widgets/jupyterlab-manager/lib/*',
                  'lib/@jupyter-widgets/jupyterlab-manager/schema/*',
                  'lib/@jupyter-widgets/output/*',
                  'lib/@jupyter-widgets/output/lib/*',
                  'lib/@jupyter-widgets/schema/v1/*',
                  'lib/@jupyter-widgets/schema/v2/*',
                  'lib/cassist/*',
                  'lib/codemirror/autocomplete-lsp/*',
                  'lib/codemirror/autocomplete-lsp/autocomplete/*',
                  'lib/codemirror/basic-setup/*',
                  'lib/codemirror/grammar/*',
                  'lib/codemirror/tooltip/*',
                  'lib/components/*',
                  'lib/components/code-panel/*',
                  'lib/components/dataframe-manager/*',
                  'lib/components/error-boundary/*',
                  'lib/components/file-manager/*',
                  'lib/components/hotkeys/*',
                  'lib/components/kernel-manager/*',
                  'lib/components/libs/*',
                  'lib/components/richoutput-panel/*',
                  'lib/components/richoutput-panel/data-panel/*',
                  'lib/components/richoutput-panel/experiment-panel/*',
                  'lib/components/richoutput-panel/model-panel/*',
                  'lib/components/richoutput-panel/result-panel/*',
                  'lib/components/richoutput-panel/summary-panel/*',
                  'lib/components/sidebar/*',
                  'lib/components/tests/*',
                  'lib/components/user-manager/*',
                  'lib/interfaces/*',
                  'pages/*',
                  'public/*',
                  'public/icons/*',
                  'redux/*',
                  'redux/actions/*',
                  'redux/reducers/*',
                  'redux/store/*',
                  'styles/*',
                  'tests/intergration/Dockerfile',
                  'tests/intergration/Dockerfile',
                  'tests/intergration/Dockerfile',
                  'tests/intergration/Dockerfile',
                  'tests/intergration/cypress.json',
                  'tests/intergration/cypress.json',
                  'tests/intergration/cypress.json',
                  'tests/intergration/cypress.json',
                  'tests/intergration/cypress/fixtures/*',
                  'tests/intergration/cypress/integration/cyc/*',
                  'tests/intergration/cypress/integration/data/*',
                  'tests/intergration/cypress/plugins/*',
                  'tests/intergration/cypress/results/*',
                  'tests/intergration/cypress/support/*',
                  'tests/intergration/old-test.js',
                  'tests/intergration/old-test.js',
                  'tests/intergration/old-test.js',
                  'tests/intergration/old-test.js',
                  'tests/intergration/package.json',
                  'tests/intergration/package.json',
                  'tests/intergration/package.json',
                  'tests/intergration/package.json',
                  'tests/unit/*',
                  'tests/unit/pages/*',
                  'theme/*'],
 'cyc_next_app.server': ['build/*', 'ls/*', 'routes/*']}

install_requires = \
['cnextlib>=0.2.0,<0.3.0',
 'jupyter-client>=7.2.2,<8.0.0',
 'jupyterlab>=3.3.4,<4.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'mlflow>=1.25.1,<2.0.0',
 'multipledispatch>=0.6.0,<0.7.0',
 'netron>=5.7.8,<6.0.0',
 'pandas==1.3.5',
 'plotly>=5.7.0,<6.0.0',
 'protobuf==3.20.1',
 'python-language-server>=0.36.2,<0.37.0',
 'requests>=2.27.1,<3.0.0',
 'send2trash>=1.8.0,<2.0.0',
 'sentry-sdk>=1.5.12,<2.0.0',
 'simplejson>=3.17.6,<4.0.0',
 'zmq>=0.0.0,<0.0.1']

entry_points = \
{'console_scripts': ['cnext-init = cyc_next_app.__main__:main',
                     'cnext-path = cyc_next_app.__main__:path',
                     'cnext-run = cyc_next_app.__main__:start']}

setup_kwargs = {
    'name': 'cnext-test',
    'version': '0.4.19',
    'description': 'The data-centric workspace for AI & DS',
    'long_description': '# CNext Instructions\n\n[Website] - [Documentation] - [Docker Image] - [Overview Video]\n\n## Setup and run the CNext workspace\n\nCNext is a data-centric workspace for DS and AI. Our workspace is meant to consolidate the most common tasks performed by data scientists and ML engineers. At a high level our workspace allows for:\n\n-   Data exploration & transformation\n-   Model development / exploration\n-   Production code generation\n-   Dashboard & App Generation\n-   Experiment Management\n\n## Features\n\n-   Interactive Python coding envrionment with native Python output (think Jupyter replacement)\n-   Smart code suggestion (categorical values and column names)\n-   Interactive data exploration\n-   Automative visualizations\n-   Experiment and model management\n-   Instant dashboarding\n\n## Installation via Docker\n\nStep 1: `Download` [cnext] sample project here and `extract`\n\nStep 2: `setup` [Docker] on your computer\n\nStep 3: Run `docker login` login with your docker account\n\nStep 4: Point destination command to `cnext` directory\n\nStep 5: Run command `docker-compose up -d`\n\n-   The web application will launch at : `http://localhost:CLIENT_PORT` or `http://127.0.0.1:CLIENT_PORT/` (CLIENT_PORT default is 3000)\n-   To stop the application: `docker-compose down`\n-   Note: Pay attention at `CLIENT_PORT`, and `SERVER_PORT` in `.env` file (you will have to change these ports if you already use them on your machine, by default they are 3000 and 4000 respectively)\n\n## Installation via Pip\n\nStep 1: `Download` [cnext] sample project here. This folder will be use as an input in Step 4\n\nStep 2: Make sure `Nodejs` is available in your computer (try `npm --version` and make sure it work)\n\nStep 3: `run` command `pip install cnext`\n\nStep 4: `run` command `cnext-init`\n\n-   Input `Enter path to the cnext sample project created in Step 1` and hit `Enter` (Example `C:/Skywalker`)\n    \u200b\n\nStep 6 `run` command `cnext-run`\n\n-   Web application will launch at : `http://localhost:CLIENT_PORT` or `http://127.0.0.1:CLIENT_PORT/` (CLIENT_PORT default is 3000)\n-   Stop application: `Ctrl + c | Command + c`\n-   Note: Pay attention at `CLIENT_PORT`, and `SERVER_PORT` in `.env` file (you will have to change these ports if you already use them on your machine, by default they are 3000 and 4000 respectively)\n\n## Release Note\n\n# v0.4.18 - 0.4.19\n\n##### NEW\n\n-   Show failed execution status marker\n\n# v0.4.14 - v0.4.16\n\n##### NEW\n\n-   Add configs for dataframe manager.\n-   Implement grouped lines folding\n\n##### IMPROVED\n\n-\n\n##### FIXED\n\n-   Fix bugs and improve stability for model visualization.\n\n# v0.4.12 - v0.4.13\n\n##### NEW\n\n-   Allow code to be executed on any file.\n-   Improve code suggestion behavior.\n-   Load config dynamically from "config.json".\n\n##### IMPROVED\n\n-\n\n##### FIXED\n\n-\n\n# v0.4.10\n\n##### NEW\n\n-   Assign result of a group to only the first line.\n-   Add support for full-html-page results using iframe.\n-   Add support for .sql and .json files.\n\n##### IMPROVED\n\n-   Improve the ability to support results with javascript. you can now run bokeh plot.\n\n##### FIXED\n\n-\n\n## License\n\nCopyright 2022 CycAI Inc.\n\u200b\nPermission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:\n\u200b\nThe above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.\n\u200b\nTHE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.\n\n[website]: https://www.cnext.io/\n[docker image]: https://hub.docker.com/r/cycai/cnext\n[documentation]: https://internal-lace-ae4.notion.site/Product-Documentation-0dd58ea1cfe14dfab3666c5ec633ae96\n[overview video]: https://youtu.be/5eWPkQIUfZw\n[cnext]: https://drive.google.com/file/d/1w4MU3nr0E14PAS_5NmruuoMSfQeE5MLK\n[docker]: https://www.docker.com/products/docker-desktop/\n',
    'author': 'CycAI Inc',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://cyc-ai.com/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9',
}


setup(**setup_kwargs)



import os
import traceback

from libs.message_handler import BaseMessageHandler
from libs.message import ContentType

from libs import logs
from libs.config import read_config
from libs.message import ProjectCommand
from project_manager import files, projects
from project_manager.interfaces import WorkspaceMetadata
from project_manager.interfaces import FileManagerMessageParams

log = logs.get_logger(__name__)


class MessageHandler(BaseMessageHandler):
    def __init__(self, p2n_queue, user_space, workspace_info: WorkspaceMetadata):
        super(MessageHandler, self).__init__(p2n_queue, user_space)
        # active_project: projects.ProjectMetadata = None
        # open_projects = workspace_info.open_projects
        # if workspace_info.active_project is not None:
        #     for project in open_projects:
        #         if workspace_info.active_project == project.id:
        #             active_project = project
        # if active_project:
        #     projects.set_active_project(active_project)

    def handle_message(self, message):
        log.info('FileManager handle message: %s %s %s' %
                 (message.command_name, message.type, message.sub_type))
        try:
            messageParams = FileManagerMessageParams(message.metadata)

            result = None
            if message.command_name == ProjectCommand.list_dir:
                result = files.list_dir(
                    messageParams.norm_project_path, messageParams.norm_path)
                type = ContentType.DIR_LIST
            elif message.command_name == ProjectCommand.get_open_files:
                result = projects.get_open_files()
                type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.set_working_dir:
                result = projects.set_working_dir(messageParams.norm_path)
                type = ContentType.NONE
            elif message.command_name == ProjectCommand.set_project_dir:
                result = projects.set_project_dir(messageParams.norm_path)
                type = ContentType.NONE
            elif message.command_name == ProjectCommand.read_file:
                result = files.read_file(messageParams.norm_project_path, messageParams.norm_path,
                                            messageParams.timestamp)
                if result == None:
                    type = ContentType.NONE
                else:
                    type = ContentType.FILE_CONTENT
            elif message.command_name == ProjectCommand.save_file:
                result = files.save_file(
                    messageParams.norm_project_path, messageParams.norm_path, content=message.content)
                type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.close_file:
                result = projects.close_file(
                    messageParams.norm_path, messageParams.open_order)
                type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.open_file:
                result = projects.open_file(
                    messageParams.norm_path, messageParams.open_order)
                type = ContentType.FILE_METADATA
            # elif message.command_name == ProjectCommand.get_active_project:
            #     result = projects.get_active_project()
            #     type = ContentType.PROJECT_METADATA
            elif message.command_name == ProjectCommand.save_state:
                result = files.save_state(
                    messageParams.norm_project_path, messageParams.norm_path, content=message.content)
                type = ContentType.FILE_METADATA
            elif message.command_name == ProjectCommand.save_project_settings:
                result = projects.save_project_settings(
                    content=message.content)
                type = ContentType.PROJECT_METADATA
            elif message.command_name == ProjectCommand.get_project_settings:
                result = projects.get_project_settings()
                type = ContentType.PROJECT_METADATA
            elif message.command_name == ProjectCommand.get_workspace_metadata:
                result = projects.get_workspace_metadata()
                type = ContentType.WORKSPACE_METADATA
            elif message.command_name == ProjectCommand.set_workspace_metadata:
                result = projects.save_workspace_metadata(message.content)
                type = ContentType.WORKSPACE_METADATA
            elif message.command_name == ProjectCommand.set_active_project:
                result = projects.set_active_project(message.content)
                type = ContentType.WORKSPACE_METADATA
            elif message.command_name == ProjectCommand.add_project:
                result = projects.add_project(message.content)
                type = ContentType.WORKSPACE_METADATA

            # create reply message
            message.type = type
            message.content = result
            message.error = False
            self._send_to_node(message)

        except:
            trace = traceback.format_exc()
            log.error("%s" % (trace))
            error_message = MessageHandler._create_error_message(
                message.webapp_endpoint, trace, command_name=message.command_name)
            self._send_to_node(error_message)
